package org.jenkinsci.plugins.p4_client.client;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.jenkinsci.plugins.p4_client.workspace.StaticWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.StreamWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.TemplateWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.Workspace;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.p4java.client.IClient;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.option.client.SyncOptions;
import com.perforce.p4java.option.server.SwitchClientViewOptions;
import com.perforce.p4java.server.IOptionsServer;

public class ConnectionHelper {

	private Logger logger = LoggerFactory.getLogger(ConnectionHelper.class);

	private final ConnectionConfig config;
	private final IOptionsServer connection;
	private IClient iclient;

	public ConnectionHelper(ConnectionConfig config) throws Exception {
		this.config = config;
		this.connection = ConnectionFactory.getConnection(config);
	}

	public boolean isConnected() {
		return connection.isConnected();
	}

	public boolean isUnicode() {
		try {
			return connection.getServerInfo().isUnicodeEnabled();
		} catch (Exception e) {
			return false;
		}
	}

	public boolean login(AuthorisationConfig auth) throws Exception {
		connection.setUserName(auth.getUsername());

		// CHARSET is not defined (only for client access)
		if (connection.getServerInfo().isUnicodeEnabled()) {
			connection.setCharsetName("utf8");
		}

		// exit early if logged in
		if (isLogin())
			return true;

		switch (auth.getType()) {
		case PASSWORD:
			String pass = auth.getPassword();
			connection.login(pass);
			break;

		case TICKET:
			String ticket = auth.getTicketValue();
			connection.setAuthTicket(ticket);
			break;

		case TICKETPATH:
			String path = auth.getTicketPath();
			connection.setTicketsFilePath(path);
			break;

		default:
			throw new Exception("Unknown Authorisation type: " + auth.getType());
		}

		return isLogin();
	}

	private boolean isLogin() throws Exception {
		String status = connection.getLoginStatus();
		if (status.contains("not necessary")) {
			return true;
		}
		if (status.contains("ticket expires in")) {
			return true;
		}
		// If there is a broker or something else that swallows the message
		if (status.isEmpty()) {
			return true;
		}

		logger.info("P4:login failed '" + status + "'");
		return false;
	}

	/**
	 * Create default client from Configuration
	 * 
	 * @param user
	 * @return
	 * @throws Exception
	 */
	public void setClient(Workspace workspace) throws Exception {

		if (config.isUnicode()) {
			String charset = "utf8";
			connection.setCharsetName(charset);
		}

		switch (workspace.getType()) {
		case STATIC:
			StaticWorkspaceImpl singletonWs = (StaticWorkspaceImpl) workspace;
			iclient = connection.getClient(singletonWs.getName());
			if (iclient == null) {
				throw new Exception("Undefined workspace: "
						+ singletonWs.getName());
			}
			break;

		case TEMPLATED:
			TemplateWorkspaceImpl templateWs = (TemplateWorkspaceImpl) workspace;
			String template = templateWs.getTemplateName();
			SwitchClientViewOptions opts;
			opts = new SwitchClientViewOptions();
			connection.switchClientView(template, templateWs.getName(), opts);
			break;

		case STREAM:
			StreamWorkspaceImpl streamWs = (StreamWorkspaceImpl) workspace;
			if (iclient == null) {
				logger.info("... Creating stream client: " + streamWs.getName());
				Client implClient = new Client();
				implClient.setName(streamWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(streamWs.getName());
			}
			iclient.setStream(streamWs.getStreamName());
			break;

		default:
			throw new Exception("Unknown workspace type: "
					+ workspace.getType());
		}

		// Ensure root and host fields are set
		iclient.setRoot(workspace.getRootPath());
		iclient.setHostName(workspace.getNodeName());

		// Save client spec
		iclient.update();

		// Set active client for this connection
		connection.setCurrentClient(iclient);
	}

	/**
	 * Sync files to workspace
	 * 
	 * @return true if new files, false if up-to-date
	 * @throws Exception
	 */
	public boolean syncFiles() throws Exception {
		// build file revision spec
		List<IFileSpec> syncFiles;
		syncFiles = FileSpecBuilder.makeFileSpecList("//...");

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		List<IFileSpec> syncMsg = iclient.sync(syncFiles, syncOpts);
		return !trapFileSpecs(syncMsg, "file(s) up-to-date.");
	}

	public void disconnect() throws Exception {
		connection.disconnect();
	}

	/**
	 * Validates returned FileSpec from operation. Will suppress warning
	 * messages in the ignore string.
	 * 
	 * @param fileSpecs
	 * @param ignore
	 * @throws ConverterException
	 */
	public void validateFileSpecs(List<IFileSpec> fileSpecs, String... ignore)
			throws Exception {
		for (IFileSpec fileSpec : fileSpecs) {
			if (fileSpec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();

				// superfluous p4java message
				boolean unknownMsg = true;
				ArrayList<String> ignoreList = new ArrayList<String>();
				ignoreList.add("file does not exist");
				ignoreList.addAll(Arrays.asList(ignore));
				for (String istring : ignoreList) {
					if (msg.contains(istring)) {
						if (logger.isTraceEnabled()) {
							logger.trace(msg);
						}
						// its a known message
						unknownMsg = false;
					}
				}

				// check and report unknown message
				if (unknownMsg) {
					logger.warn("p4java: " + msg);
				}
			}
		}
	}

	/**
	 * Look for a message in the returned FileSpec from operation.
	 * 
	 * @param fileSpecs
	 * @param ignore
	 * @return
	 * @throws ConverterException
	 */
	public boolean trapFileSpecs(List<IFileSpec> fileSpecs, String trap)
			throws Exception {
		for (IFileSpec fileSpec : fileSpecs) {
			if (fileSpec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();
				if (msg.contains(trap)) {
					return true;
				}
			}
		}
		return false;
	}
}
